import Emittery from "emittery";
import { promisify } from "util";

//needed to pass information on wether or not the promise has been aborted witout 'this'
class AbortableEmittery extends Emittery {
  public aborted = false;

  abort() {
    this.aborted = true;
    this.emit("abort");
  }
}

export default class AbortablePromise<T> implements Promise<T | null> {
  private emitter: AbortableEmittery;

  private result_promise: Promise<T | null>;
  private fulfilled = false;
  private result: T;

  constructor(
    make_generator: (
      await_or_abort: <L>(
        promise: Promise<L> | AbortablePromise<L>,
        on_abort?: () => Promise<unknown> | void
      ) => Promise<L>,
      resolve?: Function
    ) => AsyncGenerator
  ) {
    const emitter = new AbortableEmittery();

    function await_or_abort<L = any>(promise: AbortablePromise<L>): Promise<L>;
    function await_or_abort<L = any>(
      promise: Promise<L> | AbortablePromise<L>,
      on_abort?: () => Promise<unknown> | void
    ): Promise<L | null> {
      console.log("await_or_abort", promise);
      if (AbortablePromise.isAbortable(promise) && !on_abort) {
        on_abort = async () => promise.abort();
      }
      if (on_abort === undefined) {
        throw new Error(
          "on_abort is required when the first argument is not AbortablePromise"
        );
      }
      return new Promise((resolve, reject) => {
        let resolved = false;
        emitter.on("abort", () => {
          if (!resolved) {
            (on_abort as Function)();
            resolve(null);
            resolved = true;
          }
        });
        promise
          .then((result: L) => {
            if (!resolved) {
              resolve(result);
              resolved = true;
            }
          })
          .catch((err) => {
            reject(err);
          });
      });
    }
    this.result_promise = new Promise(async (resolve, reject) => {
      let step_value;
      const generator = make_generator(await_or_abort);
      const e = new Error();
      do {
        if (emitter.aborted) {
          resolve(null);
          return;
        }
        step_value = await generator.next();
      } while (!step_value.done);
      resolve(step_value.value);
      this.fulfilled = true;
      this.result = step_value.value;
    });
    this.emitter = emitter;
  }

  static isAbortable(o: unknown): o is AbortablePromise<unknown> {
    return (o as { abort: boolean }).abort;
  }

  abort() {
    this.emitter.abort();
  }

  get [Symbol.toStringTag]() {
    return "AbortablePromise";
  }

  then<TResult1 = T, TResult2 = never>(
    onfulfilled?:
      | ((value: T) => TResult1 | PromiseLike<TResult1>)
      | undefined
      | null,
    onrejected?:
      | ((reason: any) => TResult2 | PromiseLike<TResult2>)
      | undefined
      | null
  ): AbortablePromise<TResult1 | TResult2> {
    const self = this;
    return new AbortablePromise(async function* (await_or_abort) {
      if (self.fulfilled) {
        onfulfilled && onfulfilled(self.result);
        return;
      }
      const ret = await await_or_abort(self.result_promise, () => {
        self.abort();
      });
      yield;
      if (onfulfilled) {
        onfulfilled(ret as T);
      }
    });
  }

  catch<TResult = never>(
    onrejected?:
      | ((reason: any) => TResult | PromiseLike<TResult>)
      | undefined
      | null
  ): AbortablePromise<T | TResult> {
    const self = this;
    return new AbortablePromise(async function* (await_or_abort) {
      await await_or_abort(self.result_promise, () => {
        self.abort();
      }).catch(onrejected);
      yield;
    });
  }

  finally(onfinally?: (() => void) | undefined | null): AbortablePromise<T> {
    const self = this;
    return new AbortablePromise(async function* (await_or_abort) {
      await await_or_abort(self);
      onfinally && onfinally();
    });
  }

  static deadlyRace<L>(
    promises: AbortablePromise<L>[],
    should_kill_others: (arg: L) => Promise<boolean> = async () => true
  ): AbortablePromise<L> {
    return new AbortablePromise(async function* (await_or_abort) {
      return await_or_abort(
        new Promise((resolve, _) => {
          let resolved = false;
          const callback = async (arg: L) => {
            if (resolved) return;
            if (arg === null) return;
            resolved = await should_kill_others(arg);
            if (!resolved) return;
            for (const promise of promises) {
              promise.abort();
            }
            resolve(arg);
          };
          for (const promise of promises) {
            promise.then(callback);
          }
        }),
        () => {
          promises.forEach((promise) => promise.abort());
        }
      );
    });
  }
}
